/*
    zhttp.c - http client
    Copyright (C) 2012 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <zhttp.h>

#include "libziaint.h"

#include "eprintf.h"
#include <zdebug.h>
#include <zerror.h>
#include <zglib.h>
#include <zhash.h>
#include <zpath.h>
#include <zselect.h>
#include <zsock.h>
#include <zstr.h>
#include <zthread.h>

#include <string.h>

#include <glib.h>

#ifdef Z_MSC
#pragma warning(disable : 4996)
#endif

#define SLOWDOWNx 100

static void zhttp_adns_callback(struct zasyncdns *adns, int n, int *family, int *socktype, int *protocol, int *addrlen, union zsockaddr *addr, char *errorstr);
static void zhttp_connected_handler(void *arg);
static void zhttp_write_handler(void *arg);
static void zhttp_read_handler(void *arg);
static void zhttp_connecting_timer(void *arg);

struct zhttp *zhttp_init(void){
    struct zhttp *http = g_new0(struct zhttp, 1);
    http->request = zbinbuf_init();
    http->response = zbinbuf_init();    
	http->state = ZHTTPST_NEW;
	http->dataofs = -1;
	http->cookies = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	return http;
}

void zhttp_free(struct zhttp *http){
	if (!http) return;
	if (http->sock >= 0){
		zselect_set(http->zsel, http->sock, NULL, NULL, NULL, NULL);
		closesocket(http->sock);
		http->sock = -1;
	}
	zbinbuf_free(http->request);
    zbinbuf_free(http->response);
    zasyncdns_free(http->adns);
    g_free(http->errorstr);
    g_free(http->server);
    g_free(http->serveraddr);
    g_free(http->page);
	g_free(http->url);
	g_hash_table_remove_all(http->cookies);

	zhttp_post_free(http);
	g_free(http->datastr);
	if (http->connecting_timer_id > 0) zselect_timer_kill(http->zsel, http->connecting_timer_id);

	g_free(http);
}

void zhttp_post_free(struct zhttp *http){
	unsigned int i;

	if (!http->posts) return;

	for (i = 0; i < http->posts->len; i++){
		struct zhttp_post_var *var = (struct zhttp_post_var*)g_ptr_array_index(http->posts, i);
		g_free(var->name);
		g_free(var->filename);
		g_free(var->value);
		g_free(var);
	}
	g_ptr_array_free(http->posts, TRUE);
	http->posts = NULL;
}

void zhttp_disconnect(struct zhttp *http, char *errorstr){
	if (http->sock < 0) return;
	g_free(http->errorstr);
	http->errorstr = errorstr;
	if (http->errorstr) http->state = ZHTTPST_ERROR;
	zselect_set(http->zsel, http->sock, NULL, NULL, NULL, NULL);
	closesocket(http->sock);
	http->sock = -1;
	http_get_data(http);
	if (http->callback) http->callback(http);
}


static void zhttp_prepare1(struct zhttp *http, struct zselect *zsel, const char *url, void *arg){
    char *u, *c, *host;

	http->zsel = zsel;
	http->arg = arg;
	http->dataofs = -1;
	http->sent = 0;
    http->url = g_strdup(url);
	
    u = g_strdup(url);

	c = z_strcasestr(u, "https:"); // we do not support https, try http
	if (c != NULL){
		memmove(c + 4, c + 5, strlen(c + 4) + 1);
	}

    c = z_strcasestr(u, "http://");
    if (c){
        host = c + strlen("http://");
    }else{
        host = u;
    }
     
    c = strchr(host, '/');
    if (c){
        http->page = g_strdup(c);
        *c = '\0';
    }else{
        http->page = g_strdup("/");
    }

    c = strchr(host, ':');
    if (c){
        http->port = atoi(c + 1);
        *c = '\0';
    }else{
        http->port = 80;
    }
    http->server = g_strdup(host);

	zbinbuf_truncate(http->request, 0);
	zbinbuf_truncate(http->response, 0);
    g_free(u);
}

static void http_fill_cookies(gpointer key, gpointer value, gpointer user_data){
	GString *gs = (GString *)user_data;

	if (gs->len > 0) g_string_append(gs, "; ");
	g_string_sprintfa(gs, "%s=%s", (char*)key, (char*)value);
}


static void zhttp_headers(struct zhttp *http){

	zbinbuf_sprintfa(http->request, "Host: %s\r\n", http->server);
	zbinbuf_append(http->request, "User-Agent: Libzia " Z_VERSION "\r\n");
	zbinbuf_append(http->request, "Accept-Encoding:\r\n");
	zbinbuf_append(http->request, "Connection: close\r\n");
	if (g_hash_table_size(http->cookies) > 0){
		GString *cookies = g_string_new("");
		g_hash_table_foreach(http->cookies, http_fill_cookies, cookies);
		zbinbuf_sprintfa(http->request, "Cookie: %s\r\n", cookies->str);	
		g_string_free(cookies, TRUE);
	}

}

static void zhttp_do(struct zhttp *http, struct zselect *zsel, void (*callback)(struct zhttp *)){
	http->origreqlen = http->request->len;
	http->adns = zasyncdns_init();
	http->callback = callback;
	zasyncdns_getaddrinfo(
            http->adns,
            zsel,
            zhttp_adns_callback,
            http->server,
            AF_INET,
            http);
	http->state = ZHTTPST_DNS;
}

void zhttp_get(struct zhttp *http, struct zselect *zsel, const char *url, void (*callback)(struct zhttp *), void *arg){
	zhttp_prepare1(http, zsel, url, arg);
	zbinbuf_sprintfa(http->request, "GET %s HTTP/1.0\r\n", http->page);
	zhttp_headers(http);
	zbinbuf_append(http->request, "\r\n");
	zhttp_do(http, zsel, callback);
}


void zhttp_post(struct zhttp *http, struct zselect *zsel, const char *url, void (*callback)(struct zhttp *), void *arg){
	struct zbinbuf *zbb;
	char *boundary = g_strdup_printf("---------------------------%d%d%d%d", rand(), rand(), rand(), rand());

	zhttp_prepare1(http, zsel, url, arg);
	zbinbuf_sprintfa(http->request, "POST %s HTTP/1.1\r\n", http->page);
	zhttp_headers(http);
	zbinbuf_sprintfa(http->request, "Content-Type: multipart/form-data; boundary=%s\r\n", boundary);
	zbb = zbinbuf_init();
	if (http->posts){
		unsigned int i;
    	
		for (i = 0; i < http->posts->len; i++){
		    struct zhttp_post_var *var = (struct zhttp_post_var*)g_ptr_array_index(http->posts, i);
			zbinbuf_sprintfa(zbb, "--%s\r\n", boundary);
			if (var->filename != NULL) {
				char *contenttype = "application/octet-stream";
				const char *ext = z_extension(var->filename);
				if (strcasecmp(ext, ".png") == 0) contenttype = "image/png";

				zbinbuf_sprintfa(zbb, "Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"\r\n", var->name, var->filename);
				zbinbuf_sprintfa(zbb, "Content-Type: %s\r\n", contenttype);
				zbinbuf_sprintfa(zbb, "\r\n");
				if (var->localfilename != NULL)
					zbinbuf_append_file(zbb, var->localfilename);
				else
					zbinbuf_append(zbb, var->value);
				zbinbuf_sprintfa(zbb, "\r\n");
			}else{
				zbinbuf_sprintfa(zbb, "Content-Disposition: form-data; name=\"%s\"\r\n", var->name);
				zbinbuf_sprintfa(zbb, "\r\n");
				zbinbuf_sprintfa(zbb, "%s\r\n", var->value);
			}
		}
		zbinbuf_sprintfa(zbb, "--%s--\r\n", boundary);
		
	}
	zbinbuf_sprintfa(http->request, "Content-Length: %d\r\n", zbb->len);
	zbinbuf_append(http->request, "\r\n");
	zbinbuf_append_bin(http->request, zbb->buf, zbb->len);
/*    {
static int fileN = 1;
#warning TODO
        char fn[256];
        sprintf(fn, "post%d.bin", fileN++);
        zbinbuf_write_to_file(http->request, fn); 
    }*/

	zhttp_do(http, zsel, callback);
	zbinbuf_free(zbb);
	g_free(boundary); 
}

void zhttp_post_add(struct zhttp *http, const char *name, const char *value){
	
	struct zhttp_post_var *var = g_new0(struct zhttp_post_var, 1);
	var->name = g_strdup(name);
	var->value = g_strdup(value);

	if (!http->posts) http->posts = g_ptr_array_new();
	g_ptr_array_add(http->posts, var);
}
 
void zhttp_post_add_file_mem(struct zhttp *http, const char *name, const char *filename, const char *value){
	
	struct zhttp_post_var *var = g_new0(struct zhttp_post_var, 1);
	var->name = g_strdup(name);
	var->filename = g_strdup(filename);
	var->value = g_strdup(value);

	if (!http->posts) http->posts = g_ptr_array_new();
	g_ptr_array_add(http->posts, var);
}

void zhttp_post_add_file_disk(struct zhttp *http, const char *name, const char *filename, const char *localfilename){
	
	struct zhttp_post_var *var = g_new0(struct zhttp_post_var, 1);
	var->name = g_strdup(name);
	var->filename = g_strdup(filename);
	var->localfilename = g_strdup(localfilename);

	if (!http->posts) http->posts = g_ptr_array_new();
	g_ptr_array_add(http->posts, var);
}

static void zhttp_adns_callback(struct zasyncdns *adns, int n, int *family, int *socktype, int *protocol, int *addrlen, union zsockaddr *addr, char *errorstr){
    struct zhttp *http = (struct zhttp *)adns->arg;
    int i, ret;
    char errbuf[256];
	GString *gs2;            
        

    if (errorstr != NULL){
        g_free(http->errorstr);
        http->errorstr = g_strdup_printf("Can't resolve %s: %s", http->server, errorstr);
        http->callback(http);
        return;
    }

    if (n == 0){
        g_free(http->errorstr);
        http->errorstr = g_strdup_printf("Can't resolve %s: No data returned", http->server);
        http->callback(http);
        return;
    }
    
    for (i = 0; i < n; i++){
//        dbg("socket(%d, %d, %d)\n", family[i], socktype[i], protocol[i]);
        
        //http->sock = socket(family[i], socktype[i], protocol[i]);
        http->sock = socket(family[i], SOCK_STREAM, 0);
        if (http->sock < 0) continue;

        if (z_sock_nonblock(http->sock, 1)){
            closesocket(http->sock);
            http->sock = -1;
            continue;
        }
        
        switch (family[i]){
            case AF_INET:                
                ((struct sockaddr_in *)(addr + i))->sin_port = htons(http->port);                
                break;
#ifdef AF_INET6            
            case AF_INET6:              
                ((struct sockaddr_in6 *)(addr + i))->sin6_port = htons(http->port);                
                break;
#endif            
            default:                   
                closesocket(http->sock);                
                http->sock = -1;
                continue;       // unsupported protocol family        
        }       

		gs2 = g_string_new("");
		z_sock_ntoa(gs2, family[i], addr + i);
		g_free(http->serveraddr);
		http->serveraddr = g_strdup(gs2->str);
        g_string_free(gs2, TRUE);       

        ret = connect(http->sock, (struct sockaddr *)(addr + i), addrlen[i]);     
        if (ret < 0){           
            int err = z_sock_errno;         
            if (z_sock_wouldblock(err)){                
				//dbg("connecting to %s\n", http->serveraddr);      
                zselect_set(adns->zsel, http->sock, NULL, zhttp_connected_handler, NULL, http);           
				http->state = ZHTTPST_CONNECTING;
				http->connecting_timer_id = zselect_timer_new(http->zsel, 30000, zhttp_connecting_timer, http);
            }else{                
                g_free(http->errorstr);
				http->errorstr = g_strdup_printf("Can't connect to %s, %s", http->serveraddr, z_sock_strerror());                
                closesocket(http->sock);                
                http->sock = -1;         
            }           
        }        
        else{            
            zhttp_connected_handler(http);        
        }        
        break;
    }
};


static void zhttp_connecting_timer(void *arg){
	struct zhttp *http = (struct zhttp*)arg;									   
	http->connecting_timer_id = -1;
	zhttp_disconnect(http, g_strdup_printf("Timeout connecting to %s:%d", http->server, http->port));
}

static void zhttp_connected_handler(void *arg){
    struct zhttp *http = (struct zhttp *)arg;
	char errbuf[256];

	if (z_sock_error(http->sock)){
        zhttp_disconnect(http, g_strdup_printf("Can't connect to %s:%d, %s", http->server, http->port, z_sock_strerror()));
        return;
    }

	zselect_set(http->zsel, http->sock, NULL, zhttp_write_handler, NULL, http);
	http->state = ZHTTPST_REQUEST;
}

static void zhttp_write_handler(void *arg){
    struct zhttp *http = (struct zhttp *)arg;
	int ret, tosend;
	char errbuf[256];

	tosend = http->request->len;
#ifdef SLOWDOWN
	tosend = Z_MIN(tosend, SLOWDOWN);
	usleep(100000);
#endif


	ret = send(http->sock, http->request->buf, tosend, 0);
    //dbg("send(%d) = %d\n", tosend, ret);
	if (ret <= 0){
		zhttp_disconnect(http, g_strdup_printf("Error writing http socket: %s", z_sock_strerror()));
		return;
	}

	http->sent += ret;

    //dbg("buf=%d -> ", http->request->len);
	zbinbuf_erase(http->request, 0, ret);
    //dbg("%d\n", http->request->len);
	if (http->request->len == 0){
		zselect_set(http->zsel, http->sock, zhttp_read_handler, NULL, NULL, http);
		http->state = ZHTTPST_HEADER;
	}
}

#define ZHTTP_BUFSIZE 4096

static void zhttp_read_handler(void *arg){
    struct zhttp *http = (struct zhttp *)arg;
	int ret;
	char buf[ZHTTP_BUFSIZE];
	char errbuf[256];
	char *c;
	int i, ncnt, torecv;
	
	torecv = ZHTTP_BUFSIZE;
#ifdef SLOWDOWN
	torecv = Z_MIN(ZHTTP_BUFSIZE, SLOWDOWN);
	usleep(100000);
#endif

	ret = recv(http->sock, buf, torecv, 0);
	if (ret < 0){
		zhttp_disconnect(http, g_strdup_printf("Error reading http socket: %s", z_sock_strerror()));
		return;
	}
	
	if (ret == 0){
		int ofs = 0;
		
		// parse status code
		http->status = 0;
		for (i = 0; i < http->response->len; i++){
			if (http->response->buf[i] == '\n'){
				zhttp_disconnect(http, g_strdup("Can't parse status line"));				
				return;
			}
			if (http->response->buf[i] == ' '){
				char s[100];
				ofs = i + 1;
				zbinbuf_getline(http->response, &ofs, s, sizeof(s));
				http->status = atoi(s);

				
				if (http->status != 200){
					zhttp_disconnect(http, g_strdup_printf("HTTP error %s", s));				
					return;
				}
				break;
			}
			
		}

		if (!http->status){
			zhttp_disconnect(http, g_strdup("Can't parse status code"));				
			return;
		}

		http->state = ZHTTPST_DONE;
		zhttp_disconnect(http, NULL);
		return;
	}

	zbinbuf_append_bin(http->response, buf, ret);

	if (http->dataofs < 0){
		c = http->response->buf;
		ncnt = 0;
		for (i = 0; i < http->response->len; i++, c++){
			if (*c == '\r') continue;
			if (*c != '\n') {
				ncnt = 0;
				continue;
			}
			ncnt++;
			if (ncnt == 2){
				http->dataofs = i + 1;
				http->state = ZHTTPST_DATA;
				// save cookies
				zhttp_store_cookies(http, http->response->buf, i);
				break;
			}
		}
	}
}

void zhttp_status(struct zhttp *http, GString *gs){
	switch (http->state){
		case ZHTTPST_NEW:
			g_string_append(gs, "Waiting for command");
			break;
		case ZHTTPST_DNS:
			g_string_sprintfa(gs, "Resolving %s", http->server);
			break;
		case ZHTTPST_CONNECTING:
			g_string_sprintfa(gs, "Connecting to %s", http->serveraddr);
			break;
		case ZHTTPST_REQUEST:
			g_string_sprintfa(gs, "Sending request %d KB / %d KB", http->sent / 1024, http->origreqlen / 1024);
			break;
		case ZHTTPST_HEADER:
			g_string_append(gs, "Fetching headers");
			break;
		case ZHTTPST_DATA:
			g_string_printf(gs, "Downloading %d KB", (http->response->len - http->dataofs) / 1024);
			break;
		case ZHTTPST_DONE:
			g_string_append(gs, "Done");
			break;
		default:
			g_string_sprintfa(gs, "Unknown state %d", (int)http->state);
			break;
	}
	if (http->url) {
		char *c = g_strdup(http->url);
		z_strip_from(c, '?');
		g_string_sprintfa(gs, "\n%s", c);
		g_free(c);
	}
}

int zhttp_write_data(struct zhttp *http, const char *filename){
	int towrite, ret;
	char *wfn;
	FILE *f;

	towrite = http->response->len - http->dataofs;
	
	if (towrite <= 0 || http->dataofs <= 0){
		g_free(http->errorstr);
		http->errorstr = g_strdup_printf("Bad http response (len=%d ofs=%d)", http->response->len, http->dataofs);
		return -1;
	}
	wfn = g_strdup(filename);
	z_wokna(wfn);
	f = fopen(wfn, "wb");
	if (!f){
		g_free(http->errorstr);
		http->errorstr = g_strdup_printf("Can't write %s", wfn);
		g_free(wfn);
		return -2;
	}
	ret = fwrite(http->response->buf + http->dataofs, 1, towrite, f);
	if (ret != towrite){
		g_free(http->errorstr);
		http->errorstr = g_strdup_printf("Can't write to %s", wfn);
		g_free(wfn);
		return -3;
	}
	fclose(f);
	g_free(wfn);
	return 0;
}

#define SCLEN 11
void zhttp_store_cookies(struct zhttp *http, const char *data, int len){
	int i;
	const char *s, *pos;

	for (i = 0; i < len; i++){
		s = data + i;
		pos = strchr(s, '\n');
		if (pos == NULL) break;

		if (strncasecmp(s, "Set-Cookie:", SCLEN) == 0){
			char *key, *val, *oldval;
			char *d = g_strndup(s + SCLEN, (pos - s) - SCLEN);
			key = z_trim(d);
			z_strip_from(d, ';');

			val = strchr(key, '=');
			if (val != NULL){
				*val = '\0';
				val++;
				oldval = (char *)g_hash_table_lookup(http->cookies, key);
				if (oldval != NULL) g_hash_table_remove(http->cookies, key);
				g_hash_table_insert(http->cookies, g_strdup(key), g_strdup(val));
			}
			g_free(d);
		}
		i += (pos - s);
	}
}

char *http_get_data(struct zhttp *http){
	int size;
	zg_free0(http->datastr);

	if (http->dataofs <= 0)
		size = 1;
	else
		size = http->response->len - http->dataofs + 1 + 1;

	http->datastr = (char *)g_malloc(size);
	zbinbuf_getstr(http->response, http->dataofs, http->datastr, size);

	return http->datastr;
}

void zhttp_post_json(struct zhttp *http, struct zselect *zsel, const char *url, const char *json, void (*callback)(struct zhttp *), void *arg){
	
    zhttp_prepare1(http, zsel, url, arg);
	zbinbuf_sprintfa(http->request, "POST %s HTTP/1.1\r\n", http->page);
	zhttp_headers(http);
	zbinbuf_sprintfa(http->request, "Content-Type: application/json\r\n");
	zbinbuf_sprintfa(http->request, "Content-Length: %d\r\n", strlen(json));
	zbinbuf_append(http->request, "\r\n");
	zbinbuf_append_bin(http->request, json, strlen(json));
    /*{
#warning TODO
        char fn[256];
        sprintf(fn, "post%d.bin", fileN++);
        zbinbuf_write_to_file(http->request, fn, 0, http->request->len); 
        exit(-1);
    }*/ 

	zhttp_do(http, zsel, callback);
}
